%!octave 
%
%
%
%       OS : GNU/Linux 4.2.5
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: rnn_gradient.m,v 1.8 2017/02/26 18:47:56 mechanoid Exp $

% - - - - - - - - - - - - - - - - 
function retval = rnn_gradient(X,W,y),
   % процедура вычисления градиента BPTT - Backpropagation through time
   [o,s,h] = rnn_bptt_forward(X,W) ; % прямой проход
   er = rnn_bptt_backward(W,y,o,s) ; % обратный проход
   dW =  rnn_bptt_gradient(X,W,s,er,h) ; % вычисляем значения градиента для всех весов

   retval = dW ;

endfunction

% - - - - - - - - - - - - - - - - 
function [o,s,h] = rnn_bptt_forward(X,W),
   % прямой проход
   % для каждого примера:
   % последовательно вычисляем состояния скрытого слоя s для всех элементов входной последовательности
   % вычисляем выход сети о

   n = size(X,1) ; % размер одного элемента последовательности
   m = size(X,2) ; % длинна последовательности
   k = size(X,3) ; % количество примеров


   hs = [ size(W{1}{2},1), 1 ] ; % размер скрытого слоя
   ht = zeros(hs) ; % начальное значение выходов скрытого слоя 
   
   s = {} ; % история состояний скрытого слоя
   h = {} ; % история выходов скрытого слоя
   for t = 1:m,
      Xt = reshape( X(:,t,:) , [n,k] ) ; % элементы последовательностей номер t для всех примеров 
      st = W{1}{1}*Xt + W{1}{2}*ht + W{1}{3} ; % состояния скрытого слоя 
      ht = rnn_activate( st ) ; % выходы скрытого слоя
      s{t} = st ; % сохраняем историю состояний скрытого слоя
      h{t} = ht ; % сохраняем историю выходов скрытого слоя
   endfor

   o = softmax( W{2}{1} * ht + W{2}{2} ) ;

endfunction

% - - - - - - - - - - - - - - - - 
function retval = rnn_bptt_backward(W,y,o,s) ,
   % обратный проход
   % для каждого примера:
   % вычисляем ошибку выходного слоя
   % вычисляем ошибку скрытого слоя для последнего элемента входной последовательности
   % последовательно, в обратном порядке, вычисляем ошибки скрытого слоя для всех элементов входной последовательности


   % формируем "правильный" выход сети из номера класса
   d = zeros( size(o) ) ; 
   for i = 1:length(y),
      d(y(i),i) = 1 ; 
   endfor

   
   m = length(s) ; % длинна последовательности

   er{2} = o - d ; % ошибка выходного слоя

   % вычисляем ошибку скрытого слоя для всех элементов входной последовательности 
   % начиная с последнего (в обратном порядке)

   % ошибка скрытого слоя на последнем элементе входной последовательности
   er{1}{m} = transpose( W{2}{1} ) * er{2} .* rnn_activate_drv( s{end} ) ; 

   for t = m-1:-1:1, 
      % ошибка скрытого слоя на элементе входной последовательности номер t
      er{1}{t} = transpose( W{1}{2} ) * er{1}{t+1} .* rnn_activate_drv( s{t} ) ; 
   endfor

   retval = er ;

endfunction

% - - - - - - - - - - - - - - - - 
function retval = rnn_bptt_gradient(X,W,s,er,h),
   % вычисляем значения градиента для всех весов

   dW ={} ;

   dW{2}{1} = er{2} * transpose( h{end} )  ; % 3x7 =  3x100 * 100x7
   dW{2}{2} = sum(er{2},2) ;

   % размеры матриц: 7x11 , 7x7 , 3x7 

   n = size(X,1) ; % размер одного элемента последовательности
   m = size(X,2) ; % длинна последовательности
   k = size(X,3) ; % количество примеров

   t = 1 ; % выбираем первый элемент последовательности
   Xt = reshape( X(:,t,:) , [n,k] ) ; % элементы последовательностей номер t для всех примеров 


   dW{1}{1} = er{1}{t} * transpose( Xt ) ;
   dW{1}{2} = 0.0 ;
   dW{1}{3} = sum(er{1}{t},2) ; 

   for t = 2:m,
      Xt = reshape( X(:,t,:) , [n,k] ) ; % элементы последовательностей номер t для всех примеров 
      dW{1}{1} += er{1}{t} * transpose( Xt ) ;
      dW{1}{2} += er{1}{t} * transpose( h{t-1} ) ;
      dW{1}{3} += sum(er{1}{t},2) ;
   endfor

   retval = dW ;

endfunction
